
import { MmcifFormat } from '../../mol-model-formats/structure/mmcif';
import { Column, Table } from '../../mol-data/db';
import { mmCIF_Schema } from '../../mol-io/reader/cif/schema/mmcif';
import { Model } from '../../mol-model/structure';
import { ModelSymmetry } from '../../mol-model-formats/structure/property/symmetry';
import { PDBTMDescriptor } from './types';


export function registerTmDetSymmetry(pdbtmDescriptor: PDBTMDescriptor) {
    ModelSymmetry.Provider.formatRegistry.remove('mmCIF');
    const excludedChains = constructChainListFromOperations(pdbtmDescriptor);
    ModelSymmetry.Provider.formatRegistry.add('mmCIF',  function(model: Model) {
        return tmDetSymmetryFromMmCif(model, excludedChains);
    });
}

function constructChainListFromOperations(pdbtmDescriptor: PDBTMDescriptor): string[] {
    const excludedChains: string[] = [];
    // add chain deletes
    const biomatrix = pdbtmDescriptor.additional_entry_annotations.biomatrix;
    if (biomatrix != undefined && biomatrix.chain_deletes != undefined) {
        biomatrix.chain_deletes.forEach(
            chainId => excludedChains.push(chainId)
        );
    }
    // exclude result of transformations
    if (biomatrix != undefined && biomatrix.matrix_list != undefined) {
        biomatrix.matrix_list.forEach(
            matrix => matrix.apply_to_chain_list.forEach(
                applyItem => excludedChains.push(applyItem.new_chain_id)
            )
        );
    }

    return excludedChains;
}

function tmDetSymmetryFromMmCif(model: Model, chainsForOperExpression: string[]) {
    if (!MmcifFormat.is(model.sourceData)) return;

    let data = model.sourceData.data.db;
    let asymIdColumnData: string[][] = [];
    // update chains (asym) for each author_defined_assemly
    // NOTE: We assume that the first assembly is an author defined assembly.
    const currentAssemblyId: string = data.pdbx_struct_assembly_gen.assembly_id.value(0);
    const result: string[][] = createAsymIdColumnData(
        data.pdbx_struct_assembly_gen, currentAssemblyId, chainsForOperExpression
    );
    asymIdColumnData = asymIdColumnData.concat(result);

    const asym_id_list_column = Column.ofStringListArray(asymIdColumnData);
    // create table with new column
    let updated_pdbx_struct_assembly_gen = Table.ofColumns(
        data.pdbx_struct_assembly_gen._schema,
        {
            assembly_id: data.pdbx_struct_assembly_gen.assembly_id,
            asym_id_list: asym_id_list_column,
            oper_expression: data.pdbx_struct_assembly_gen.oper_expression
        }
    );

    return ModelSymmetry.fromData({
        symmetry: data.symmetry,
        cell: data.cell,
        struct_ncs_oper: data.struct_ncs_oper,
        atom_sites: data.atom_sites,
        pdbx_struct_assembly: data.pdbx_struct_assembly,
        pdbx_struct_assembly_gen: updated_pdbx_struct_assembly_gen,
        pdbx_struct_oper_list: data.pdbx_struct_oper_list
    });
}

function createAsymIdColumnData(pdbx_struct_assembly_gen: Table<mmCIF_Schema['pdbx_struct_assembly_gen']>, assemblyId: string, chains: string[]) {

    const asym_id_list: string[][] = [];
    for (let i = 0; i < pdbx_struct_assembly_gen._rowCount; i++) {
        const currentAsymIdList = pdbx_struct_assembly_gen.asym_id_list.value(i);
        asym_id_list.push(
            (assemblyId == pdbx_struct_assembly_gen.assembly_id.value(i))
                ? minus(currentAsymIdList, chains)
                : currentAsymIdList
        );
    }
    return asym_id_list;
}

// difference of two string arrays (interpreted as sets)
function minus(a: string[], b: string[]): string[] {
    const b_set = new Set(b);
    return a.filter(x => !b_set.has(x));
}
