import { MolScriptBuilder as MS } from '../../mol-script/language/builder';
import { PluginUIContext } from '../../mol-plugin-ui/context';
import { Mat4, Vec3 } from '../../mol-math/linear-algebra';
import { PDBTMDescriptor, PDBTMTransformationMatrix, PMS } from './types';
import { createStructureRepresentationParams } from '../../mol-plugin-state/helpers/structure-representation-params';
import { StateTransforms } from '../../mol-plugin-state/transforms';
import { StateObjectRef, StateBuilder } from '../../mol-state';
import { Expression } from '../../mol-script/language/expression';
import { MembraneOrientation } from './prop';

export async function applyTransformations(ctx: PluginUIContext, pdbtmDescriptor: PDBTMDescriptor) {
    const annotations = pdbtmDescriptor.additional_entry_annotations;

    console.log('BIOMX', annotations.biomatrix);
    if (annotations.biomatrix == undefined) {
        return;
    }

    annotations.biomatrix.matrix_list.forEach(function(mx) {
        mx.apply_to_chain_list.forEach(function(chainPair) {
            let id  = chainPair.chain_id;
            let newId = chainPair.new_chain_id;
            console.log(`before RED ${id} -> ${newId}`);
            red(ctx, mx.transformation_matrix, id);
            console.log("after RED");
        });
    });
}

/**
 * Util/test function. Preparation for chain-level operations.
 * Sets color of specified chain to red.
 *
 * @param plugin UI context
 * @param transformationMatrix 4x4 matrix describes transformation and translation
 * @param chainId Id of chain to be selected for transformation
 */
 export function red(plugin: PluginUIContext, transformationMatrix: PDBTMTransformationMatrix, chainId: string): void {
    // TODO: const color: Color = Color.fromArray([255, 0, 0], 0);
    const query: Expression = getChainExpression(chainId);

    let transformation = transformationForStateTransform(transformationMatrix);
    const structure: StateObjectRef<PMS> = plugin.managers.structure.hierarchy.current.models[0].structures[0].cell;
    const update: StateBuilder.To<any, any> = plugin.build().to(structure);

    update
        .apply(
            StateTransforms.Model.StructureSelectionFromExpression,
            { label: chainId, expression: query }
        )
        .apply(StateTransforms.Model.TransformStructureConformation, {
            "transform": { name: "matrix", params: { data: transformation, transpose: false } }
        })
        .apply(
            StateTransforms.Representation.StructureRepresentation3D,
            createStructureRepresentationParams(plugin, update.selector.data, {
                type: 'cartoon',
                color: 'chain-id',
                // TODO: Can we copy colors of parent asymmetry?
                // colorParams: { value: color }
            })
        );
    update.commit();
}

function vadd(a: Vec3, b: Vec3): Vec3 {
    return Vec3.add(Vec3.zero(), a, b);
}

function vneg(u: Vec3): Vec3 {
    return Vec3.negate(Vec3.zero(), u);
}

export function createMembraneOrientation(pdbtmDescriptor: PDBTMDescriptor): MembraneOrientation {
    const annotations = pdbtmDescriptor.additional_entry_annotations;
    const membrane = annotations.membrane;
    const transformation: Mat4 = transformationForStateTransform(membrane.transformation_matrix);
    const translation: Vec3 = Mat4.getTranslation(Vec3.zero(), transformation);

    // TODO: console.log("MEMBRANE TRNASFORMATION:\n\n", Mat4.makeTable(transformation));
    Vec3.negate(translation, translation);
    Mat4.setTranslation(transformation, translation);
    // TODO: console.log("MEMBRANE TRNASFORMATION: NEGATED trnltn:\n\n", Mat4.makeTable(transformation));

    // transpose rotation of transformation (inverse rotation)
    Mat4.transpose(transformation, Mat4.extractRotation(transformation, transformation));
    // TODO: console.log("MEMBRANE TRANSPOSED:\n\n", Mat4.makeTable(transformation));

    // transform membrane normal
    const membraneNormal: Vec3 = Vec3.fromObj(membrane.normal);
    Vec3.transformMat4(membraneNormal, membraneNormal, transformation);

    const result: MembraneOrientation = {
        planePoint1: vadd(membraneNormal, translation),
        planePoint2: vadd(vneg(membraneNormal), translation),
        centroid: translation,
        normalVector: membraneNormal,

        // (NOTE: the TMDET extension calculates and sets it during applying preset)
        radius: membrane.radius
    };
    return result;
}

export function transformationForStateTransform(tmatrix: PDBTMTransformationMatrix): Mat4 {
    // matrix expected in column-major order
    const mx: Mat4 = Mat4.fromArray(Mat4.zero(),
        [
            tmatrix.rowx.x, tmatrix.rowy.x, tmatrix.rowz.x, 0,
            tmatrix.rowx.y, tmatrix.rowy.y, tmatrix.rowz.y, 0,
            tmatrix.rowx.z, tmatrix.rowy.z, tmatrix.rowz.z, 0,
                         0,              0,              0, 1
        ], 0
    );
    Mat4.setTranslation(mx, Vec3.create(
        tmatrix.rowx.t,
        tmatrix.rowy.t,
        tmatrix.rowz.t
    ));
    // TODO: console.log('MAT4\n', Mat4.makeTable(mx));
    return mx;
}

export function getAtomGroupExpression(chainId: string, auth_array: number[]): Expression {
    // TODO console.log('auth_array:', auth_array);
    const query: Expression =
        MS.struct.generator.atomGroups({
            'residue-test': MS.core.set.has([MS.set( ...auth_array ), MS.ammp('auth_seq_id')]),
            'chain-test': MS.core.rel.eq([chainId, MS.ammp('label_asym_id')])
        });
    return query;
}

export function getChainExpression(chainId: string): Expression {
    const query: Expression =
        MS.struct.generator.atomGroups({
            'chain-test': MS.core.rel.eq([chainId, MS.ammp('label_asym_id')])
        });
    return query;
}
